//
// Fur example for OpenGL
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	"libExt.h"

#include	<glut.h>
#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Torus.h"

Vector3D	eye   ( 6, 5, 6 );			// camera position
Vector3D	light ( 5, 0, 4 );			// light position
unsigned	decalMap;
float	 	angle = 0;
Torus		torus ( 1, 3, 30, 30 );
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
unsigned	furTex [16];

/////////////////////////////////////////////////////////////////////////////////
/*
inline	float	clamp ( float x )
{
	if ( x < 0 )
		return 0;

	if ( x > 1 )
		return 1;

	return x;
}

unsigned BuildFurLightingTexture()
{
	unsigned	id;

										// size of texture
	int sSize = 16;
	int tSize = 32;
	int rSize = 16;

										// exponents for diffuse and specular
	float diffPower = 6;
	float specPower = 32;

	byte * pixels = new byte [2*sSize*tSize*rSize];
	byte * p      = pixels;

	for ( int k = 0; k < rSize; k++ )
		for ( int j = 0; j < tSize; j++ )
			for ( int i = 0; i < sSize; i++ )
			{
				float ln = float(i)/(sSize-1); 			// s = (L dot N) ^ 2
				float hn = float(j)/(tSize-1); 			// t = (H dot N) ^ 2
				float lz = sqrt (float(k)/(rSize-1));	// r = sqrt (L.z ^ 2) = abs ( l.z)

				float diffuse  = 0.2 + 0.2 * max ( 0, lz ) + 0.6 * pow ( 1 - ln, 0.5*diffPower );
				float specular = pow ( 1 - hn, 0.5*specPower );
				float shadow   = pow ( cos ( M_PI/2 * clamp(-0.2-1.4*lz)), 4 );

				*p++ = byte( 0xFF * clamp ( diffuse ) );
				*p++ = byte( 0xFF * clamp ( specular*shadow ) );
			}

	glGenTextures ( 1, &id );
	glBindTexture ( GL_TEXTURE_3D_EXT, id );

	glTexImage3DEXT ( GL_TEXTURE_3D_EXT, 0, GL_LUMINANCE8_ALPHA8, sSize, tSize, rSize, 0,
					  GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, pixels );

	glTexParameteri ( GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri ( GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri ( GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S,     GL_CLAMP_TO_EDGE );
	glTexParameteri ( GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T,     GL_CLAMP_TO_EDGE );
	glTexParameteri ( GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, GL_CLAMP_TO_EDGE );

	return id;
}
*/
void init ()
{
	glClearColor ( 0.0, 0.0, 0.0, 1.0 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );

	glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
	glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
	glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glPushMatrix ();

                               // draw torus itself
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glEnable           ( GL_TEXTURE_2D );
	glBindTexture      ( GL_TEXTURE_2D, decalMap );

	glMatrixMode   ( GL_MODELVIEW );
	glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

	glEnable   ( GL_CULL_FACE );
	glCullFace ( GL_BACK );

	torus.draw  ();

								// now start drawing shells
	float	scale     = 0.5;
	float	shadowMin = 0.2;
	float	shadowMax = 0.5;

	glDisable      ( GL_CULL_FACE );
	glEnable       ( GL_BLEND );
	glBlendFunc    ( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glMatrixMode   ( GL_TEXTURE );
	glLoadIdentity ();
	glScalef       ( 7, 12, 7 );
	glMatrixMode   ( GL_MODELVIEW );

								// draw shells from innermost to outermost
	for ( int i = 0; i < 16; i++ )
	{
		float	t      = (float) i / 16.0;
		float	shadow = shadowMin * (1 - t) + shadowMax * t;

		glBindTexture ( GL_TEXTURE_2D, furTex [i] );
		glColor4f     ( 1, 1, 1, shadow );

		torus.draw ( t * scale );
	}

	glDisable ( GL_BLEND );

	glPopMatrix     ();
	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					0, 0, 0,				// center
					0, 0, 1 );				// up
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )		//quit requested
    	exit ( 0 );
}

void	animate ()
{
	angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

	light.x = 2*cos ( angle );
	light.y = 2*sin ( angle );
	light.z = 3 + 0.3 * sin ( angle / 3 );

										// setup data
	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 500, 500 );


								// create window
	int	win = glutCreateWindow ( "OpenGL fur demo 2 (single texture only)" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init           ();
	initExtensions ();

	decalMap = createTexture2D ( true,  "../Textures/block.bmp" );

								// load fur textures
	for ( int i = 0; i < 16; i++ )
	{
		char	name [1218];

		sprintf ( name, "Textures/fur-map2-00.png" );

		furTex [i] = createTexture2D ( true, name );
	}

	glutMainLoop ();

	return 0;
}
