#
# script to build a set of fur textures and normal maps
#
# Author: Alex V. Boreskoff

import PIL, Image
import math, random

size         = 256					# size of initial texture (before resizing)
density      = 0.55
curliness    = 1.0
eps          = 0.3
numLayers    = 16
tangentScale = 0.3
numParticles = round ( density * size * size )
binStep      = 16
numBins1     = size / binStep
numBins2     = numBins1 * numBins1
bins         = []

for i in range ( numBins2 ):
	bins.append ( [] )

class Particle:
	def __init__ ( self ):
		self.x      = random.random () * ( size - 1 );
		self.y      = random.random () * ( size - 1 );
		self.length = 1     #0.8 + 0.2 * random.random ()
		self.a1     = random.random () * math.pi * 4
		self.a2     = random.random () * math.pi * 4
		self.xm1    = 2 * (random.random () - 0.5) * curliness
		self.ym1    = 2 * (random.random () - 0.5) * curliness
		self.xm2    = 2 * (random.random () - 0.5) * curliness
		self.ym2    = 2 * (random.random () - 0.5) * curliness

	def posAt ( self, time ):
		xNew = self.x + self.xm1 * math.sin ( self.a1 * time ) + self.xm2 * math.sin ( self.a2 * time )
		yNew = self.y + self.ym1 * math.sin ( self.a1 * time ) + self.ym2 * math.sin ( self.a2 * time )

		return (xNew, yNew)

	def tangentAt ( self, time ):
		dx  = self.xm1 * self.a1 * math.cos ( self.a1 * time ) + self.xm2 * self.a2 * math.cos ( self.a2 * time )
		dy  = -self.ym1 * self.a1 * math.sin ( self.a1 * time ) - self.ym2 * self.a2 * math.sin ( self.a2 * time )
		dx  = dx * tangentScale
		dy  = dy * tangentScale
		len = math.sqrt ( dx * dx + dy * dy + 1 )

		return ( dx / len, dy / len, 1 / len )

	def distanceTo ( self, other ):
		return abs ( self.x - other.x ) + abs ( self.y - other.y )

	def overlaps ( self, particles, count ):
		b = bins [self.getBin ()]

		for p in b:
			if self.distanceTo ( p ) < eps:
				return 1

		return 0

	def getBin ( self ):
		i = int ( self.x / binStep )
		j = int ( self.y / binStep )

		return i * numBins1 + j

im   = Image.new ( "RGBA", (size, size) )				# colors and alpha map
im2  = Image.new ( "RGB",  (size, size) )				# tangent map
im3  = Image.new ( "RGB",  (size, size) )				# for filtered tangent map

			# init to (black,0)
def clearImage ( image ):
	for x in range ( size ):
		for y in range ( size ):
			image.putpixel ( (x,y), (0, 0, 0, 0) )

			# do "blurring of tangent maps"
			# we use only non-black pixels (to create real map without holes)
def blurTangentMap ( map, out ):

			# for every texel calc average on non-empty pixels
	for x in range ( size ):
		for y in range ( size ):
			sum   = [0,0,0]
			count = 0
			for dx in range ( -1, 2 ):
				for dy in range ( -1, 2 ):
					x1 = (x + dx + size) % size
					y1 = (y + dy + size) % size
					c  = map.getpixel ( (x1,y1) )

					if c [0] != 0 or c [1] != 0 or c [2] != 0:
						count   = count + 1
						sum [0] = sum [0] + c [0]
						sum [1] = sum [1] + c [1]
						sum [2] = sum [2] + c [2]

					# now average them and map to [0,1]
			if count > 0:
					sum [0] = sum [0] / (count)
					sum [1] = sum [1] / (count)
					sum [2] = sum [2] / (count)

#			v       = ( 2*sum [0] - 1, 2*sum [1] - 1, 2*sum [2] - 1)
#			length  = math.sqrt ( v [0] * v [0] + v [1] * v [1] + v [2] * v [2] )
#			if length > 0:

			out.putpixel ( (x,y), (round ( sum [0] ), round ( sum [1] ), round ( sum [2] ) ) )

							# now create a set of particles
print "Creating", numParticles, " particles"

particles = []

for i in range ( numParticles ):
	while 1:
		p = Particle ()
		b = p.getBin ()

		if not p.overlaps ( particles, i ):
			particles.append ( p )
			bins [b].append  ( p )
			break

							# start particle simulation
for layer in range ( numLayers ):
	t = float ( layer ) / float ( numLayers )

	clearImage ( im  )
	clearImage ( im2 )

	for i in range ( numParticles ):
		if particles [i].length < t:		# length check
			continue

		pos = particles [i].posAt ( t )		# get position at
		x   = round ( pos [0] ) % size		# do wrapping
		y   = round ( pos [1] ) % size
		im.putpixel ( (x, y ), (255, 170, 64, 255) )

		tang  = particles [i].tangentAt ( t )
		red   = 128 + round ( 127 * tang [0] )
		green = 128 + round ( 127 * tang [1] )
		blue  = 128 + round ( 127 * tang [2] )

		im2.putpixel ( (x, y), (red, green, blue) )

											# here we should do "normal blur" of tangent map
	blurTangentMap ( im2, im3 )
	blurTangentMap ( im3, im2 )

	fileName  = "fur-map-%02d.png"         % ( layer )
	fileName2 = "fur-tangent-map-%02d.png" % ( layer )

	im. save ( fileName, "png" )
	im2.save ( fileName2, "png" )

	print "Layer %d done" % (layer)

